#include <X11/Intrinsic.h>
#include <stdio.h>
#include <math.h>
#include "particle.h"

#define TABRANGE	512
#define PI		3.14159265

static double sinTable[TABRANGE], cosTable[TABRANGE], sphereTable[TABRANGE];

static Particle *FreeParticles = NULL;

/*
*** Allocate a new particle structure.
 */
static Particle *AllocParticle()
{
    Particle *part = FreeParticles;

    if (part==NULL) {
	part = (Particle *)malloc(sizeof(Particle));
    } else {
	FreeParticles = part->next;
    }
    return part;
}

/*
*** CreateAtFixedRate: Create babies at a fixed rate
 */
static void CreateFixedRate(sys)
System *sys;
{
    Particle *part;
    int c = sys->start_c + sys->start_n;
    int d = sys->start_d;

    while (c >= d) {
	part = AllocParticle();
	part->next = sys->babies;

	part->dx = 0;
	part->dy = 0;

	sys->babies = part;
	c -= d;
    }
    sys->start_c = c;
}

/*
*** CreateRandom
*** Create babies randomly, at some rate
 */
static void CreateRandom(sys)
System *sys;
{
    Particle *part;
    int ran = RRANGE(0,sys->start_d+sys->start_c);
    int n = sys->start_n;
    int d = sys->start_d;

    while (ran > d) {
	part = AllocParticle();
	part->next = sys->babies;

	part->dx = 0;
	part->dy = 0;

	sys->babies = part;
	ran -= d;
    }
}


static void CreateBabiesFromTimeout(sys, deadpart)
System *sys;
Particle *deadpart;
{
    Particle *part;
    int c = sys->start_c + sys->start_n;
    int d = sys->start_d;
    int x = deadpart->x;
    int y = deadpart->y;
    int dx = deadpart->dx;
    int dy = deadpart->dy;


    while (c > d) {
	part = AllocParticle();
	part->next = sys->babies;

	part->x = x;
	part->y = y;
	part->dx = dx;
	part->dy = dy;

	sys->babies = part;
	c -= d;
    }
    sys->start_c = c;
}

static void PositionXY(sys)
System *sys;
{
    Particle *part = sys->babies;
    int x = sys->pos.x1;
    int y = sys->pos.y1; 

    while (part) {
	part->x = x;
	part->y = y;
	part = part->next;
    }
}

/*
*** DirectDXDY: Set the initial velocity to a fixed dx/dy value
 */
static void DirectDXDY(sys)
System *sys;
{
    Particle *part;
    int dx = sys->vel.x1;
    int dy = sys->vel.y1;

    for (part = sys->babies; part; part = part->next) {
	part->dx = dx;
	part->dy = dy;
    }
}

/*
*** DirectRTheta: Set the initial velocity to a fixed R/Theta value
 */
static void DirectRTheta(sys)
System *sys;
{
    Particle *part;

    for (part = sys->babies; part; part = part->next) {
	double r = sys->vel.r1;
	int theta = sys->vel.t1;

	part->dx = (int)(r*cosTable[theta]);
	part->dy = (int)(r*sinTable[theta]);
    }
}

/*
*** DirectInBox:
*** Set the initial velocity within a linear range of dx/dy values
 */
static void DirectInBox(sys)
System *sys;
{
    Particle *part = sys->babies;

    while (part) {
	part->dx = RRANGE(sys->vel.x1, sys->vel.x2);
	part->dy = RRANGE(sys->vel.y1, sys->vel.y2);
	part = part->next;
    }
}

/*
*** DirectCircular:
*** Set the initial velocity within a linear range of R/Theta values
 */
static void DirectCircular(sys)
System *sys;
{
    Particle *part;

    for (part = sys->babies; part; part = part->next) {
	double r = RRANGE(sys->vel.r1, sys->vel.r2);
	int theta = RRANGE(sys->vel.t1, sys->vel.t2);

	part->dx = (int)(r*cosTable[theta]);
	part->dy = (int)(r*sinTable[theta]);
    }
}

/*
*** DirectSpherical:
*** Set the initial velocity within a range of R/Theta values with a
*** R distribution designed to give a spherical effect!
 */
static void DirectSpherical(sys)
System *sys;
{
    Particle *part = sys->babies;
    double rbase = RRANGE(sys->vel.r1, sys->vel.r2);

    while (part) {
	double r = rbase*sphereTable[RRANGE(0, TABRANGE-1)];
	int theta = RRANGE(sys->vel.t1, sys->vel.t2);
		
	part->dx = (int)(r*cosTable[theta]);
	part->dy = (int)(r*sinTable[theta]);
	part = part->next;
    }
}

/*
*** ReDirectSpherical:
*** Change the velocity within a range of R/Theta values with a
*** R distribution designed to give a spherical effect!
 */
static void RedirectSpherical(sys)
System *sys;
{
    Particle *part = sys->babies;

    while (part) {
	double r = sys->vel.x2*sphereTable[RRANGE(0, TABRANGE-1)];
	int theta = RRANGE(sys->vel.y1, sys->vel.y2);
		
	part->dx += (int)(r*cosTable[theta]);
	part->dy += (int)(r*sinTable[theta]);
	part = part->next;
    }
}

/*
*** StartTimeout
*** Set the timeout of this particle within the given range
 */
static void StartTimeout(sys)
System *sys;
{
    Particle *part = sys->babies;

    while (part) {
	part->timeout = RRANGE(sys->time_min, sys->time_max);
	part = part->next;
    }
}

/*
*** MoveLinear
 */
static void MoveLinear(sys)
System *sys;
{
    Particle *part = sys->adults;

    while (part) {
	part->x += part->dx;
	part->y += part->dy;

	part = part->next;
    }
}

/*
*** Accelerate
 */
static void Accelerate(sys)
System *sys;
{
    Particle *part = sys->adults;
    int ddx = sys->ddx;
    int ddy = sys->ddy;

    while (part) {
	part->dx += ddx;
	part->dy += ddy;

	part = part->next;
    }
}

/*
*** Drag
 */
static void Drag(sys)
System *sys;
{
    Particle *part = sys->adults;
    int friction = FRAC(1L) - sys->friction;

    while (part) {
	part->dx = INT(part->dx*friction);
	part->dy = INT(part->dy*friction);

	part = part->next;
    }
}

/*
*** Gravitate
*** Only apply virtical acceleration, for efficieny
 */
static void Gravitate(sys)
System *sys;
{
    Particle *part = sys->adults;
    int ddy = sys->ddy;

    while (part) {
	part->dy += ddy;

	part = part->next;
    }
}

#ifdef XXX
static void ClipToSpace(sys)
System *sys;
{
    Particle *corpses = NULL;
    Particle *prev = NULL;
    Particle *part = sys->adults;
    Space *space;
    Space *nspace;
    float bounce = -sys->bounce;

/*
 * This looks more complicated and slow than it really is!
 */
    while (part) {
	space = part->space;

	if (part->y < space->box.y1) {
	    nspace = space->down;
	} else if (part->x < space->box.x1) {
	    nspace = space->left;
	} else if (part->x >= space->box.x2) {
	    nspace = space->right;
	} else if (part->y >= space->box.y2) {
	    nspace = space->up;
	} else {
/*
 * Particle is still in this box, so move on to the next
 */
	    prev = part;
	    part = part->next;
	    continue;
	}
/*
 * Particle moved into a new box. Work out what we should do next..
 */
	switch ((int)nspace) {
	case (int)DIE:
	    if (prev) {
		prev->next = part->next;
		part->next = corpses;
		corpses = part;
		part = prev->next;
	    } else {
		sys->adults = part->next;
		part->next = corpses;
		corpses = part;
		part = sys->adults;
	    }
	    break;
	case (int)BOUNCEX:
	    part->x  -= part->dx;
	    part->dx *= bounce;
	    break;
	case (int)BOUNCEY:
	    part->y  -= part->dy;
	    part->dy *= bounce;
	    break;
	default:
	    part->space = nspace;
	}
    }
    sys->corpses = corpses;
}
#endif /* XXX */

static void ClipToBox(sys)
System *sys;
{
    Particle *prev = NULL;
    Particle *part = sys->adults;
    Particle *corpses = sys->corpses;
    int x1 = sys->clip.x1;
    int y1 = sys->clip.y1;
    int x2 = sys->clip.x2;
    int y2 = sys->clip.y2;

    while (part) {
	if (part->x < x1 || part->y < y1 ||
	    part->x >= x2 || part->y >= y2) {
	    if (prev) {
		prev->next = part->next;
		part->next = corpses;
		corpses = part;
		part = prev->next;
	    } else {
		sys->adults = part->next;
		part->next = corpses;
		corpses = part;
		part = sys->adults;
	    }
	} else {
	    prev = part;
	    part = part->next;
	}
    }
    sys->corpses = corpses;
}

static void CheckTimeout(sys)
System *sys;
{
    Particle *prev = NULL;
    Particle *part = sys->adults;
    Particle *corpses = sys->corpses;

    while (part) {
	if (--part->timeout == 0) {
	    if (prev) {
		prev->next = part->next;
		part->next = corpses;
		corpses = part;
		part = prev->next;
	    } else {
		sys->adults = part->next;
		part->next = corpses;
		corpses = part;
		part = sys->adults;
	    }
	} else {
	    prev = part;
	    part = part->next;
	}
    }
    sys->corpses = corpses;
}

/*
*** Reproduce:
*** Each adult produces a set of babies
 */
static void Reproduce(sys)
System *sys;
{
    Particle *part;
    System *sys2 = sys->baby_system;

    for (part = sys->adults; part; part = part->next) {
	CreateBabiesFromTimeout(sys2, part);
    }
}

/*
*** Explode:
*** Each corpse produces a whole new set of babies
 */
static void Explode(sys)
System *sys;
{
    Particle *part;
    System *sys2 = sys->new_system;

    for (part = sys->corpses; part; part = part->next) {
	CreateBabiesFromTimeout(sys2, part);
    }
}

/*
*** Reincarnate:
*** Each corpse produces a new baby in a (different) system
 */
static void Reincarnate(sys)
System *sys;
{
    Particle *part = sys->corpses;

    if (part) {
	System *sys2 = sys->new_system;

	while (part->next) {
	    part = part->next;
	}

	part->next = sys2->babies;
	sys2->babies = sys->corpses;
	sys->corpses = NULL;
    }
}

static void DisplaySystem(anim, sys)
Animation *anim;
System *sys;
{
    Particle *prev=0;
    Particle *part;
    XPoint *fore_points = sys->fore_points;
    XPoint *back_points = sys->back_points;
    XPoint *fore = fore_points;
    XPoint *tmp;
    int yoff = anim->win_height - 1;

/*
 * Un-draw particles
 */
    if (sys->n_back) {
	XDrawPoints(anim->dpy, anim->win, anim->gc_back,
		    back_points, sys->n_back, CoordModeOrigin);
    }
/*
 * Remove corpses
 */
    if ((part = sys->corpses) != 0) {
	do  {
	    prev = part;
	    part = part->next;
	} while (part);
	prev->next = FreeParticles;
	FreeParticles = sys->corpses;
	sys->corpses = 0;
    }

/*
 * Draw live particles
 */
    part = sys->adults;

    while (part) {
	fore->x = INT(part->x);
	fore->y = yoff - INT(part->y);
	fore++;
	part = part->next;
    }
/*
 * Mature Babies
 */
    if ((part = sys->babies)!=0) {
	do {
	    fore->x = INT(part->x);
	    fore->y = yoff - INT(part->y);
	    fore++;
	    prev = part;
	    part = part->next;
	} while (part);
	prev->next = sys->adults;
	sys->adults = sys->babies;
	sys->babies = 0;
    }
/*
 * Do the graphics itself
 */
    if (fore > fore_points) {
	XDrawPoints(anim->dpy, anim->win, sys->gc,
		    fore_points, fore-fore_points, CoordModeOrigin);
    }

    tmp = sys->fore_points;
    sys->fore_points = sys->back_points;
    sys->back_points = tmp;
    sys->n_back = fore - fore_points;
}

void DoAnimation(anim)
Animation *anim;
{
    System *sys = anim->systems;

    while (sys) {
	ProcList *pl = sys->procs->next;

	while (pl) {
	    (*pl->proc.proc)(sys);
	    pl = pl->next;
	}
	sys = sys->next;
    }

    sys = anim->systems;

    while (sys) {
	DisplaySystem(anim, sys);

	sys = sys->next;
    }
}

Animation *InitAnimation(iSys)
System *iSys;
{
    Animation *anim;
    int i;

    for (i=0; i<TABRANGE; i++) {
	float angle = (i-(TABRANGE/2))*PI/((float)TABRANGE/2.0);

	sinTable[i] = sin(angle);
	cosTable[i] = cos(angle);
	sphereTable[i] = sin(acos(i/(float)(TABRANGE-1)));
    }

    anim = (Animation *)malloc(sizeof(Animation));

    anim->systems = iSys;

    return anim;
}

ProcList *CreateProcList()
{
    ProcList *list = (ProcList *)malloc(sizeof(ProcList));

    list->next = NULL;

    return list;
}

void AddToProcList(list, name, proc)
ProcList *list;
char *name;
Procedure proc;
{
    ProcList *new = (ProcList *)malloc(sizeof(ProcList));

    new->proc.name = name;
    new->proc.proc = proc;

    while (list->next) list = list->next;

    new->next = list->next;
    list->next = new;

    return;
}

void AddParticleProcs(list)
ProcList *list;
{
    AddToProcList(list, "CreateFixedRate",	CreateFixedRate);
    AddToProcList(list, "CreateRandom",		CreateRandom);
    AddToProcList(list, "PositionXY",		PositionXY);
/*    AddToProcList(list, "PositionInBox",	PositionInBox);*/
/*    AddToProcList(list, "PositionInCircle",	PositionInCircle);*/
/*    AddToProcList(list, "PositionInSphere",	PositionInSphere);*/
    AddToProcList(list, "DirectDXDY",		DirectDXDY);
    AddToProcList(list, "DirectRTheta",		DirectRTheta);
    AddToProcList(list, "DirectInBox",		DirectInBox);
    AddToProcList(list, "DirectCircular",	DirectCircular);
    AddToProcList(list, "DirectSpherical",	DirectSpherical);
    AddToProcList(list, "RedirectSpherical",	RedirectSpherical);
    AddToProcList(list, "StartTimeout",		StartTimeout);

    AddToProcList(list, "MoveLinear",		MoveLinear);
    AddToProcList(list, "Accelerate",		Accelerate);
    AddToProcList(list, "Gravitate",		Gravitate);
    AddToProcList(list, "Drag",			Drag);
    AddToProcList(list, "ClipToBox",		ClipToBox);
    AddToProcList(list, "CheckTimeout",		CheckTimeout);
    AddToProcList(list, "Reproduce",		Reproduce);

    AddToProcList(list, "Explode",		Explode);
    AddToProcList(list, "Reincarnate",		Reincarnate);
}
