#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#ifdef SYNC
#include <X11/extensions/sync.h>
#endif /* SYNC */
#include "particle.h"
#include "setup.h"

static Display *animDpy=0;

static unsigned long sync_event;
static unsigned long sync_error;

static XtAppContext appC;
static Widget top, pane, pic, quit, go;
static Animation *anim;

typedef struct _Globals {
    unsigned long timestep;
    String system_types;
    int buffer_size;
#ifdef SYNC
    Boolean sync;
    int batch;
#endif /* SYNC */
} Globals;

static int default_buffer_size = 2048;
static Globals globals;

static XtResource globalResources[] = {
#ifdef SYNC
{"sync", "Sync", XtRBoolean, sizeof(Boolean),
     XtOffset(Globals *, sync), XtRString, (caddr_t)"Yes"},
{"batch", "Batch", XtRInt, sizeof(int),
     XtOffset(Globals *, batch), XtRString, (caddr_t)"5"},
#endif /* SYNC */
{"timestep", "TimeStep", XtRInt, sizeof(int),
     XtOffset(Globals *, timestep), XtRString, (caddr_t)"1000"},
{"systemTypes", "SystemTypes", XtRString, sizeof(String),
     XtOffset(Globals *, system_types), XtRString, (caddr_t)"default"},
{"bufferSize", "BufferSize", XtRInt, sizeof(int),
     XtOffset(Globals *, buffer_size), XtRInt, (caddr_t)&default_buffer_size}
};

static XrmOptionDescRec options[] = {
#ifdef SYNC
{ "-sync",	"*sync",		XrmoptionNoArg,		"Yes"},
{ "-nosync",	"*sync",		XrmoptionNoArg,		"No"},
#endif /* SYNC */
{ "-batch",	"*batch",		XrmoptionSepArg,	NULL},
{ "-timestep",	"*timestep",		XrmoptionSepArg,	NULL},
{ "-buffersize","*bufferSize",		XrmoptionSepArg,	NULL}
};

static void DoOneFrame()
{
    DoAnimation(anim);
}

/*ARGSUSED*/
static void DoTimeout(client_data, id)
XtPointer client_data;
XtIntervalId id;
{
    XtAppAddTimeOut(appC, globals.timestep, DoTimeout, client_data);

    DoOneFrame();
    XFlush(animDpy);
}

#ifdef SYNC

static Alarm alarm = 0;
static Counter servertime = 0;
static CARD64 sync_time;
static CARD64 timestep64;
static int frames_to_do = 0;

/*ARGSUSED*/
static void AnimTimeout(client_data, id)
XtPointer client_data;
XtIntervalId id;
{
    XWaitCondition cond;
    int i;
    
    cond.counter = servertime;
    cond.wait_type = SyncAbsolute;
    cvtLongToC64(cond.event_threshold, 1L<<20);
    C64Add(sync_time, sync_time, timestep64);
    cond.wait_value = sync_time;

    XSyncAwait(animDpy, &cond, 1);
    DoOneFrame();

    if (--frames_to_do <= 0) {
	XFlush(animDpy);
    } else {
/*
 * Add a timeout with no delay. With any luck, other events will be handled
 * before we get in again
 */
	XtAppAddTimeOut(appC, 0, AnimTimeout, 0);
    }
}

/*ARGSUSED*/
static Boolean AnimWorkProc(client_data, id)
XtPointer client_data;
XtWorkProcId id;
{
    XWaitCondition cond;
    int i;
    
    cond.counter = servertime;
    cond.wait_type = SyncAbsolute;
    cvtLongToC64(cond.event_threshold, 1L<<20);
    C64Add(sync_time, sync_time, timestep64);
    cond.wait_value = sync_time;

    XSyncAwait(animDpy, &cond, 1);
    DoOneFrame();

    if (--frames_to_do <= 1) {
	if (frames_to_do < 1) {
	    fprintf(stderr, "%d frames_to_do\n", frames_to_do);
	}
	XFlush(animDpy);
	return True;
    }

    return False;
}

static void AlarmEvent(ev)
XEvent *ev;
{
    XAlarmNotifyEvent *aev = (XAlarmNotifyEvent *)ev;

    switch (aev->state) {
    case AlarmActive:
	break;
    case AlarmExhausted:
	printf("Alarm Exhausted\n");
	return;
    case AlarmDestroyed:
	printf("Alarm Destroyed\n");
	return;
    }
/*
 * We need to keep the user-interface going so we calculate new frames
 * in the "Background".
 */

    frames_to_do += globals.batch;
    XtAppAddWorkProc(appC, AnimWorkProc, 0);
}

static int InitSYNC()
{
    Display *dpy = XtDisplay(top);
    SystemCounterList list;
    unsigned long n;
    int sync_major, sync_minor;
    unsigned long alarm_mask;
    XSyncAlarmAttributes alarm_values;

    if (!XSyncQueryExtension(dpy, &sync_event, &sync_error)) {
	fprintf(stderr, "Failed to find SYNC Extension\n");
	return 0;
    }
    if (!XSyncInfo(dpy, &sync_major, &sync_minor, &list, &n)) {
	fprintf(stderr, "Failed to get SYNC info\n");
	return 0;
    }
    printf("SYNC version %d.%2.2d\n", sync_major, sync_minor);
    if (sync_major != SYNC_MAJOR_VERSION ||
	sync_minor != SYNC_MINOR_VERSION) {
	fprintf(stderr,
		"Warning: Program version compiled for SYNC %d.%2.2d, Server is SYNC %d.%2.2d\n",
		sync_major, sync_minor,
		SYNC_MAJOR_VERSION, SYNC_MAJOR_VERSION);
    }

    servertime = 0;
    while (n-->0) {
	if (strcmp(list[n].name, "SERVERTIME")==0) {
	    servertime = list[n].counter;
	}
    }
    if (servertime==0) {
	fprintf(stderr, "Failed to find SERVERTIME counter\n");
	return 0;
    }

    alarm_mask = CAEvents;
    alarm_values.events = 0;
    alarm = XSyncCreateAlarm(dpy, alarm_mask, &alarm_values);

    cvtLongToC64(timestep64, globals.timestep);

    return 1;
}
#endif /* SYNC */

static void StartAnimation()
{
    Display *dpy = XtDisplay(top);
#ifdef SYNC
    unsigned long alarm_mask;
    XSyncAlarmAttributes alarm_values;

    if (globals.sync) {
	alarm_mask = CAEvents | CACounter | CAValue | CAIncrement | CALimit | CALimitType;

	alarm_values.events = 1;
	alarm_values.counter = servertime;

	XSyncQueryCounter(dpy, servertime, &alarm_values.value);
	cvtLongToC64(alarm_values.increment, globals.timestep*globals.batch);
	cvtLongToC64(alarm_values.limit, (unsigned long)600000);
	alarm_values.limit_type = SyncRelative;

	XSyncChangeAlarm(dpy, alarm, alarm_mask, &alarm_values);

	sync_time = alarm_values.value;
	C64Add(sync_time, sync_time, alarm_values.increment);
    } else {
#endif /* SYNC */
	XtAppAddTimeOut(appC, globals.timestep, DoTimeout, 0);
#ifdef SYNC
    }
#endif /* SYNC */

    XtVaSetValues(go,
		  XtNsensitive, 0,
		  NULL);
}

static void Quit()
{
    exit(0);
}

static void Go()
{
    StartAnimation();
}

main(argc, argv)
Cardinal argc;
char *argv[];
{
    Widget systems;
    unsigned colour;
    Display *dpy;

    top = XtVaAppInitialize(&appC, "Xparticle",
			    options, XtNumber(options),
			    &argc, argv, NULL,
			    NULL);
    if (argc != 1) {
	fprintf(stderr, "Bad args\n");
	exit(1);
    }

    InitSetup(appC);

    XtVaGetApplicationResources(top, (XtPointer)&globals,
				globalResources, XtNumber(globalResources),
				NULL);

    dpy = XtDisplay(top);
    if (globals.buffer_size > 0 &&
	globals.buffer_size != dpy->bufmax - dpy->buffer) {
	char *tmp = dpy->buffer;

	fprintf(stderr, "Changing Xlib buffer size from %d to %d\n",
		dpy->bufmax - dpy->buffer,
		globals.buffer_size);
	XFlush(dpy);
	dpy->buffer = (char *)malloc(globals.buffer_size);
	if (dpy->buffer) {
	    XFree(tmp);
	    dpy->bufmax = dpy->buffer + globals.buffer_size;
	    dpy->bufptr = dpy->buffer;
	} else {
	    fprintf(stderr, "Failed to allocate new Xlib buffer\n");
	    dpy->buffer = tmp;
	}
    }

#ifdef SYNC
    if (globals.sync) {
	if (!InitSYNC()) {
	    globals.sync = False;
	}
    }
#endif

    pane = XtVaCreateManagedWidget("pane", panedWidgetClass,
				   top,
				   NULL);
    quit = XtVaCreateManagedWidget("quit", commandWidgetClass,
				   pane,
				   NULL);
    XtAddCallback(quit, XtNcallback, Quit, NULL);
    go = XtVaCreateManagedWidget("go", commandWidgetClass,
				 pane,
				 NULL);
    XtAddCallback(go, XtNcallback, Go, NULL);

    systems = XtVaCreateManagedWidget("systems", panedWidgetClass,
				      pane,
				      NULL);

/*
 * A blank widget for our animation...
 */
    pic = XtVaCreateManagedWidget("pic", coreWidgetClass,
				  pane,
				  NULL);


/*
 * Set up the particle system
 */

    anim = DoSetup(top, systems, pic, globals.system_types);

    XtRealizeWidget(top);

/*
 * Open a second display connection for the animation itself.
 * This allows the user-interface to keep running while there are
 * animation commands waiting.
 */
    animDpy = XOpenDisplay(XDisplayName((char *)XtDisplay(pic)));
    if (animDpy==0) {
	fprintf(stderr, "Failed to re-open Display %s\n",
		XDisplayName((char *)XtDisplay(pic)));
	return 1;
    }

    anim->dpy = animDpy;
    anim->win = XtWindow(pic);
    anim->gc_back = XCreateGC(anim->dpy, anim->win, 0, NULL);
    XtVaGetValues(pic,
		  XtNbackground, &colour,
		  XtNheight, &anim->win_height,
		  NULL);
    XSetForeground(anim->dpy, anim->gc_back, colour);

#ifdef SYNC
    if (globals.sync) {
/*
 * We need to catch AlarmNotify events, and I'm not sure if that is possible
 * with a Toolkit event handler.
 */
	int notify = sync_event + AlarmNotify;

	for (;;) {
	    XEvent ev;

	    XtAppNextEvent(appC, &ev);
	    if (ev.type == notify) {
		AlarmEvent(&ev);
	    } else {
		XtDispatchEvent(&ev);
	    }
	}
    } else {
#endif /* SYNC */
	XtAppMainLoop(appC);
	return 0;
#ifdef SYNC
    }
#endif /* SYNC */
}
